﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CognitiveScience
{
    class Tree
    {
        double latitude, longitude;
        public string name = null;
        List<Tree> children = null;
        public Tree parent = null;

        public Tree(string name, double latitude, double longitude)
        {
            this.parent = null;
            this.children = new List<Tree>();
            this.name = name;
            this.latitude = latitude;
            this.longitude = longitude;
        }

        public double euclidean(Tree A)
        {        
            return Math.Sqrt((latitude - A.latitude) * (latitude - A.latitude) + (longitude - A.longitude) * (longitude - A.longitude));
        }

        public Tree giveRegion(string name)
        {
            if (name.Equals(this.name))
                return this.parent;
            foreach (Tree c in children)            
                c.giveRegion(name);
            return null;
        }

        public Tree giveNode(string name)
        {
            if (name.Equals(this.name))
                return this;
            foreach (Tree c in children)
            {
                Tree loc = c.giveNode(name);
                if (loc != null)
                    return loc;
            }
            return null;
        }

        public List<string> placesInRegion(List<string> places)
        {            
            List<string> placesInCurrent = new List<string>();
            foreach (string t in places)
            {
                if (isChild(t))
                    placesInCurrent.Add(t);
            }
            return placesInCurrent;
        }

        public bool isChild(string place)
        {
            foreach (Tree child in children)
            {
                if(child.name.Equals(place))
                    return true;
            }
            return false;
        }


        public bool isInRegion(List<string> names)
        {
            //returns true if some name are in region
            foreach (string name in names)
            {
                if (this.name.Equals(name))
                    return true;
            }

            foreach (Tree child in children)
            {
                if(child.isInRegion(names))
                    return true;
            }
            return false;
        }


        public int numberInRegion(List<string> names)
        {
            //returns true if some name are in region
            int number=0;
            foreach (string name in names)
            {
                if (this.name.Equals(name))
                    number++;
            }

            foreach (Tree child in children)
            {
                number += child.numberInRegion(names);
            }
            return number;
        }

        
        public Tree findRegionCoarseToGrain(Tree present, List<string> toGo)
        {
            Tree tmp=present;
            while (!tmp.isInRegion(toGo))
                tmp = tmp.parent;
            //go to children of tmp with minimum clusterSize
            List<Tree> ctmp = tmp.children;
            double[] distance =new double[ctmp.Count];
            for (int iter = 0; iter < ctmp.Count; iter++)
            {
                if (ctmp[iter].isInRegion(toGo))                
                    distance[iter] = Math.Sqrt((present.latitude - ctmp[iter].latitude) * (present.latitude - ctmp[iter].latitude) + (present.longitude - ctmp[iter].longitude) * (present.longitude - ctmp[iter].longitude));                
                else distance[iter] = Double.PositiveInfinity;
            }
            double min = distance.Min();
            for (int i = 0; i < ctmp.Count; i++)
            {
                if (distance[i] == min)
                    return ctmp[i];
            }
            return null;
        }

        public Tree findRegionClusteringMethod(Tree present, List<string> toGo)
        {
            Tree tmp = present;
            while (!tmp.isInRegion(toGo))
                tmp = tmp.parent;
            //go to children of tmp with minimum clusterSize
            List<Tree> ctmp = tmp.children;
            double[] clusterSize = new double[ctmp.Count];
            for (int iter = 0; iter < ctmp.Count; iter++)
            {
                int cluster=ctmp[iter].numberInRegion(toGo);                
                clusterSize[iter] = cluster;                
            }

            double max = clusterSize.Max();
            for (int i = 0; i < ctmp.Count; i++)
            {
                if (clusterSize[i] == max)
                    return ctmp[i];
            }
            return null;
        }



        public List<Tree> walkShort(Tree present, Tree region,List<string> names)
        {
            List<Tree> inRegion = new List<Tree>();
            List<Tree> answer = new List<Tree>();
            foreach (string name in names)
            {
                if (region.isInRegion(new List<string>(){name}))
                    inRegion.Add(Program.currentTree.giveNode(name));
            }

            //find clusterSize to each of this region
            Tree tmp = present;
            while(inRegion.Count!=0)
            {
                double[] distance=new double[inRegion.Count];

                for (int iter = 0; iter < inRegion.Count; iter++)
                    distance[iter] = tmp.euclidean(inRegion[iter]);

                double minD = distance.Min();
                int minIndex=-1;
                for (int iter = 0; iter < inRegion.Count; iter++)
                {
                    if (distance[iter] == minD)
                        minIndex = iter;
                }
                tmp = inRegion[minIndex];
                answer.Add(tmp);
                inRegion.RemoveAt(minIndex);
            }

            return answer;
        }

        public double featureDistance(Tree present, List<Tree> location)
        {
            double d = present.euclidean(location[0]);
            for (int iter = 0; iter < location.Count - 1; iter++)
                d = location[iter].euclidean(location[iter+1]);
            return d;
        }

        public void insert(string name, double latitude, double longitude)
        {
            Tree t = new Tree(name,latitude,longitude);
            this.children.Add(t);
            t.parent = this;
        }

        public Tree merge(List<string> places)
        {
            List<Tree> t = new List<Tree>();
            double avgLatitude = 0, avgLongitude = 0;
            Tree parent=null;
            foreach (string s in places)
            {
                Tree tmp = this.giveNode(s);
                parent=tmp.parent;
                avgLatitude += tmp.latitude;
                avgLongitude += tmp.longitude;
                t.Add(tmp);
            }

            //create a node with children as List t
            Tree newNode = new Tree("region",avgLatitude/places.Count,avgLongitude/places.Count);
            newNode.parent = parent;
            parent.children.Add(newNode);
            foreach (Tree c in t)
            {
                newNode.children.Add(c);
                c.parent = newNode;
                parent.children.Remove(c);
            }

            return newNode;
        }

        public Tree merge2(List<Tree> places)
        {
            List<Tree> t = places;
            double avgLatitude = 0, avgLongitude = 0;
            Tree parent = null;
            foreach (Tree tmp in places)
            {                
                parent = tmp.parent;
                avgLatitude += tmp.latitude;
                avgLongitude += tmp.longitude;                
            }

            //create a node with children as List t
            Tree newNode = new Tree("region", avgLatitude / places.Count, avgLongitude / places.Count);
            newNode.parent = parent;
            foreach (Tree c in t)
            {
                newNode.children.Add(c);
                c.parent = newNode;
            }

            return newNode;
        }

        public Tree merge3(string name)
        {
            Tree u=null;
            if (this.parent != null)
            {
                u = this.parent.giveNode(name);
            }
            else u  = this.giveNode(name);
            Tree z = new Tree("region",(u.latitude+this.latitude)/2,(u.longitude+this.longitude)/2);
            z.parent = this.parent;            
            this.parent.children.Remove(u);
            this.parent.children.Remove(this);
            this.parent.children.Add(z);
            u.parent = z;
            this.parent = z;            
            z.children.Add(this);
            z.children.Add(u);
            return z;
        }

        public void display()
        {
            Console.Write(this.name+" { ");
            foreach (Tree c in this.children)            
                c.display();            
            Console.Write(" }");
        }

    }
}
